@@ -22,28 +22,32 @@ class KafkaCrypto(KafkaCryptoBase):
         kp (KafkaProducer): Pre-initialized KafkaProducer, ready for
                             handling crypto-keying messages. Should
                             not be used elsewhere, as this class
-                            changes some configuration values.
-        kc (KafkaConsumer): Pre-initialized KafkaConsumer, ready ofor      
+                            changes some configuration values. Always
+                            closed on close.
+        kc (KafkaConsumer): Pre-initialized KafkaConsumer, ready for
        	       	       	    handling crypto-keying messages. Should
                             not be used elsewhere, as this class
-                            changes some configuration values.
+                            changes some configuration values. Always
+                            closed on close.
          config (str,file): Filename or File IO object in which
                             configuration data is stored. Set to None
                             to load from the default location based
                             on nodeID. Must be seekable, with read/
                             write permission, honor sync requests,
                             and not be written by any other program.
+                            Only closed if opened by us.
            cryptokey (obj): Optional object implementing the
                             necessary public/private key functions
                             (get/sign_spk,get/use_epk,
                             wrap/unwrap_opaque).
                             Set to None to load from the default
                             location in the configuration file.
-                seed (obj): Optional object implementing the 
+                seed (obj): Optional object implementing the
                             necessary ratchet functions
                             (increment, generate). Set to None to
                             load from the default location in
-                            the configuration file.
+                            the configuration file. Only closed
+                            if opened by us.
   """
 
   #
@@ -86,7 +90,9 @@ class KafkaCrypto(KafkaCryptoBase):
       if (seed is None):
         seed = self._nodeID + ".seed"
         self._cryptostore.store_value('ratchet', 'file#' + seed)
+    self._seed_close = False
     if (isinstance(seed,(str,))):
+      self._seed_close = True
       seed = Ratchet(file=seed)
     if (not hasattr(seed, 'increment') or not inspect.isroutine(seed.increment) or not hasattr(seed, 'get_key_value_generators') or not inspect.isroutine(seed.get_key_value_generators)):
       raise KafkaCryptoError("Invalid seed source supplied!")
@@ -133,14 +139,34 @@ class KafkaCrypto(KafkaCryptoBase):
             # stored pgens do not have generators as they should never be used for active production
             # (but secret stays around so lost consumers can catch up)
 
+    self._mgmt_thread_stop = False
     self._mgmt_thread = Thread(target=self._process_mgmt_messages,daemon=True)
     self._mgmt_thread.start()
 
+  def close(self):
+    self._mgmt_thread_stop = True
+    try:
+      self._mgmt_thread.join()
+    except:
+      pass
+    finally:
+      self._mgmt_thread = None
+    try:
+      if self._seed_close:
+        self._seed.close()
+    except:
+      pass
+    finally:
+      self._seed = None
+    super().close()
+
   # Main background processing loop. Must assume that it can "die" at any
   # time, even mid-stride, so ordering of operations to ensure atomicity and
   # durability is critical.
   def _process_mgmt_messages(self):
     while True:
+      if self._mgmt_thread_stop:
+        break
       # First, process messages
       # we are the only thread ever using _kc, _kp, so we do not need the lock to use them
       msgs = self._kc.poll(timeout_ms=self.MGMT_POLL_INTERVAL, max_records=self.MGMT_POLL_RECORDS)
